/*
	APL Compiler
	
	Run time routines - Memory Allocation / TRS manipulation
	tim budd
*/

/*
	The APL Compiler is Public Domain
	It may be freely redistributed as long as this notice of authorship
	is retained with the sources.

		Tim Budd
		Oregon State University
		Department of Computer Science
		Corvallis, Oregon 97331 USA

	No guarantees are given as to the suitablity of this code for any
	purposes whatsoever

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 02111-1307, USA.

*/

/* to get buffersize: */
#include "run.h"

#define NILTRS  (struct trs_struct *) 0

/* turn on for vectalloc prints */
#define VDEBUG 0

/* some globals */
/*  integer vector 1, scalar one shape */
int aplc_ivone[1] = {1};
/* integer zero - vector for shape */
int aplc_ivzero[1] = {0};

/* declare global trs for null args of fns */
trs_t aplc_nulltrs = {0,0,(int *)0,0,0};

/* local function declarations */
/*static void aplc_link_scalar(struct trs_struct *res);*/

#if VDEBUG 
/* type strings    - must match type list in aplc.h */
static char *types[] = { "unknown type", 
			 "bit", "bool", "label", "int",
			 "real", "complex", "quaternion", "octonion",
			 "char", "function", "operator", "boxed", 
			 "any" };
/* sws  string representation of type */
extern char *str_type_name(int type);

extern char *
str_type_name(int type)
{
  return(types[type]);
}
#endif /* #if VDEBUG  */

/* null fn for dummy operator arguments */
extern int 
aplc_nullf(struct trs_struct *r, struct trs_struct *a, 
	   struct trs_struct *b)
{ return 0;}


/* assign - assign a trs structure to an identifier 
            note this 
	    copies type, rank
	    does not copy shape or value, but just assigns pointers  */
extern void
aplc_assign(struct trs_struct * id, struct trs_struct * trs)
{
  id->type = trs->type;
  id->rank = trs->rank;
  id->shape = trs->shape;
  id->size = trs->size;
  id->alloc_ind = trs->alloc_ind;
  id->scalar = trs->scalar;
  if (id->alloc_ind & APLC_ALLOC_VAL_S) {
    /* scalar */
    if (id->size > 1)
      aplc_error("[aplc_assign] internal scalar error");
    aplc_link_scalar(id);
  } else
    id->value = trs->value;

  /* now we don't want 2 freeable trs 
     - turn original off */
  RBIT(trs->alloc_ind, APLC_ALLOC_SHAPE_F);
  RBIT(trs->alloc_ind, APLC_ALLOC_VAL_F); 

  return;
}

/*  equate values, with some type checking and changing.
    can't just increase storage of id here - this
    must be done more carefully
*/
/* sws ...
   assign_ind - assign part of a trs structure to part of an identifier
                id[idi]= trs[trsi]
*/
extern void
aplc_assign_ind(struct trs_struct * id, int idi, 
		struct trs_struct * trs, int trsi)
{
  int itype, ttype;

  /* assume id is of rank >0 check rank of trs, treat scalar
     seperately check types
  */
  itype = id->type;
  ttype = trs->type;

  if (trs->rank == 0) {
    /* scalar right term - ignore index */
    switch (itype) {
    case APLC_CHAR:
      if (ttype != APLC_CHAR) {
	aplc_error("[aplc_assign_ind] impossible conversion to APLC_CHAR");
	break;
      } else {
	id->value.cp[idi] = *(trs->value.cp);
	break;
      }
    case APLC_BOOL:
    case APLC_INT:
      if (ttype != APLC_INT) {
	aplc_error("[aplc_assign_ind] impossible conversion to APLC_INT");
	break;
      } else {
	id->value.ip[idi] = *(trs->value.ip);
	break;
      }
    case APLC_REAL:
      if (ttype == APLC_INT) {
	id->value.rp[idi] = *(trs->value.ip);
	break;
      } else if (ttype == APLC_REAL) {
	id->value.rp[idi] = *(trs->value.rp);
	break;
      } else {
	aplc_error("[aplc_assign_ind] impossible conversion to REAL");
	break;
      }
    default:
      fprintf(aplcerr, "type %d\n", itype);
      aplc_error("[aplc_assign_ind] impossible type");
    }
  } else {
    /* expected case, right index necessary */
    switch (itype) {
    case APLC_CHAR:
      if (ttype != APLC_CHAR) {
	aplc_error("[aplc_assign_ind] impossible conversion to APLC_CHAR");
	break;
      } else {
	id->value.cp[idi] = trs->value.cp[trsi];
	break;
      }
    case APLC_BOOL:
    case APLC_INT:
      if (ttype != APLC_INT) {
	aplc_error("[aplc_assign_ind] impossible conversion to APLC_INT");
	break;
      } else {
	id->value.ip[idi] = trs->value.ip[trsi];
	break;
      }
    case APLC_REAL:
      if (ttype == APLC_INT) {
	id->value.rp[idi] = trs->value.ip[trsi];
	break;
      } else if (ttype == APLC_REAL) {
	id->value.rp[idi] = trs->value.rp[trsi];
	break;
      } else {
	aplc_error("[aplc_assign_ind] impossible conversion to REAL");
	break;
      }
    default:
      fprintf(aplcerr, "type %d\n", itype);
      aplc_error("[aplc_assign_ind] impossible type");
    }
  }
}

/* sws
   initialize a trs structure - to zilde
   */
extern void
aplc_inittrs(struct trs_struct * trs)
{
  trs->type = APLC_UKTYPE;
  /* scalar 
     trs->rank = 0;
     trs->shape = (int *) 0;*/
  trs->rank = 1;
  trs->shape = aplc_ivzero;
  trs->size = 0; 
  trs->alloc_ind = APLC_UNALLOC;
  trs->alloc_type = APLC_UKTYPE;
  trs->alloc_rank = 0;
  trs->alloc_size = 0;
  trs->value.ip = (int *) 0;
  return;
}

/* aplc_mp2res - get a value from storage for a variable 
   - take it from a mp_struct and put it in a res_struct */
extern void
aplc_mp2res(union res_struct * res, union mp_struct * mp, int i, int type)
{
  int j, i2;

  switch (type) {
    case APLC_CHAR:
    res->c = *(mp->cp + i);
    break;
  case APLC_BOOL:
  case APLC_INT:
    res->i = *(mp->ip + i);
    break;
  case APLC_REAL:
    res->r = *(mp->rp + i);
    break;
  case APLC_COMPLEX:
    i2 = 2*i;
    res->z[0] = mp->rp[i2];
    res->z[1] = mp->rp[1+i2];
    break;
  case APLC_QUAT:
    i2 = 4*i;
    res->q[0] = mp->rp[i2];
    res->q[1] = mp->rp[1+i2];
    res->q[2] = mp->rp[2+i2];
    res->q[3] = mp->rp[3+i2];
    break;
  case APLC_OCT:
    i2 = 8*i;
    for (j=0; j<8; j++)
      res->o[j] = mp->rp[j+i2];
    break;
  case APLC_BOXED:
    res->trs = mp->trsp[i];
    break;
  case APLC_FNP:
    res->fn = mp->fnp[i];
    break;
  default:
    fprintf(aplcerr, "type %d\n", type);
    aplc_error("[aplc_mp2res] impossible condition");
  }
}

/* setmp - get a value into storage for a variable 
   - take it from a res_struct and put it in a mp_struct 
   - presumably the mp_struct is from a trs */
extern void
aplc_setmp(union res_struct *res, union mp_struct *mp, int i, int type)
{
  int j, i2;

  switch (type) {
    case APLC_CHAR:
    *(mp->cp + i) = res->c;
    break;
  case APLC_BOOL:
  case APLC_INT:
    mp->ip[i] = res->i;
    break;
  case APLC_REAL:
    mp->rp[i] = res->r;
    break;
  case APLC_COMPLEX:
    i2 = 2*i;
    mp->rp[i2] = res->z[0];
    mp->rp[1+i2] = res->z[1];
    break;
  case APLC_QUAT:
    i2 = 4*i;
    mp->rp[0+i2] = res->q[0];
    mp->rp[1+i2] = res->q[1];
    mp->rp[2+i2] = res->q[2];
    mp->rp[3+i2] = res->q[3];
    break;
  case APLC_OCT:
    i2 = 8*i;
    for (j=0; j<8; j++)
      mp->rp[j+i2] = res->o[j];
    break;
  case APLC_BOXED:
    /*mp->trsp[i] = res->trs;*/
    aplc_copytrs(&mp->trsp[i], &res->trs);
    break;
  case APLC_FNP:
    mp->fnp[i] = res->fn;
    break;
  default:
    fprintf(aplcerr, "type %d\n", type);
    aplc_error("[aplc_setmp] impossible condition");
  }
}

/* setup the value of a trs by linking to a res struct */
extern void
aplc_link_res(trs_t *t, res_t *res)
{
  switch(t->type) {
  case APLC_BOOL:
  case APLC_INT:
    t->value.ip = &res->i;
    break;
  case APLC_REAL:
    t->value.rp = &res->r;
    break;
  case APLC_CHAR:
    t->value.cp = &res->c;
    break;
  case APLC_COMPLEX:
    t->value.rp = &res->z[0];
    break;
  case APLC_QUAT:
    t->value.rp = &res->q[0];
    break;
  case APLC_OCT:
    t->value.rp = &res->o[0];
    break;
  default:
    aplc_error("[aplc_link_res] scalar error");
    break;
  }
  t->alloc_ind |= APLC_ALLOC_NF;
  t->alloc_type = APLC_UKTYPE;
  t->alloc_rank = 0;
  t->alloc_size = 0;
  return;
}

/* setup the value of a trs by linking to it's scalar part */
extern void
aplc_link_scalar(struct trs_struct *res)
{
  switch(res->type) {
  case APLC_BOOL:
  case APLC_INT:
    res->value.ip = &res->scalar.i;
    break;
  case APLC_REAL:
    res->value.rp = &res->scalar.r;
    break;
  case APLC_CHAR:
    res->value.cp = &res->scalar.c;
    break;
  case APLC_COMPLEX:
    res->value.rp = &res->scalar.z[0];
    break;
  case APLC_QUAT:
    res->value.rp = &res->scalar.q[0];
    break;
  case APLC_OCT:
    res->value.rp = &res->scalar.o[0];
    break;
  case APLC_FNP:
    res->value.fnp = &res->scalar.fn;
    break;
  case APLC_OPP:
    res->value.opp = &res->scalar.op;
    break;
  default:
    fprintf(stderr,"(type %d)\n", res->type);
    aplc_error("[aplc_link_scalar] scalar error");
    break;
  }
  RBIT(res->alloc_ind, APLC_ALLOC_VAL_F);
  res->alloc_ind |= APLC_ALLOC_VAL_NF;
  res->alloc_ind |= APLC_ALLOC_VAL_S;
  res->alloc_type = res->type;
  res->alloc_rank = res->rank;
  res->alloc_size = res->size;
  return;
}

/* sws
   copy an entire trs to another */
extern void
aplc_copytrs(struct trs_struct *res, struct trs_struct *in)
{
  int i,k,i2;

  /* type, rank, shape */
  aplc_settrs(res, in->type, in->rank, in->shape);

  /* values */
  if (in->alloc_ind & APLC_ALLOC_VAL_S) {
    /* scalar, linked to the values */ 
    res->scalar = in->scalar;
    aplc_link_scalar(res);
    /*printf("[copytrs] in->scalar.i %d, res->scalar.i %d\n", in->scalar.i, res->scalar.i); */
    return;
  }

  if (in->size <= 0) {
    res->alloc_type = 0;
    res->alloc_rank = 0;
    res->alloc_size = 0;
    return;
  }
  aplc_vectalloc(&res->value, in->size, in->type);
  res->alloc_ind |=  APLC_ALLOC_VAL_F;

  switch (in->type) {
  case APLC_BOXED:
    /* the value is boxed - need to recurse */
    for (i=0; i<in->size; i++)
      aplc_copytrs( &res->value.trsp[i], &in->value.trsp[i] );
    break;
  default: /* flat */
    switch(in->type) {
    default:
    case APLC_UNDEF:
    case APLC_UKTYPE:
    case APLC_BOXED:
      fprintf(stderr,"(type %d)\n", in->type);
      in->type = 1; 
      aplc_error("[aplc_copytrs] impossible case");
      break;
    case APLC_BOOL:
    case APLC_INT:
      for (i=0; i<in->size; i++)
	res->value.ip[i] = in->value.ip[i];  
      break;
    case APLC_REAL:
      for (i=0; i<in->size; i++)
	res->value.rp[i] = in->value.rp[i];  
      break;
    case APLC_COMPLEX:
      for (i=0,i2=0; i<in->size; i++,i2+=2) {
	for (k=0; k<2; k++)
	  res->value.rp[k+i2] = in->value.rp[k+i2];
      }
      break;
    case APLC_QUAT:
      for (i=0,i2=0; i<in->size; i++,i2+=4) {
	for (k=0; k<4; k++)
	  res->value.rp[k+i2] = in->value.rp[k+i2];
      }
      break;
    case APLC_OCT:
      for (i=0, i2=0; i<in->size; i++,i2+=8) {
	for (k=0; k<8; k++)
	  res->value.rp[k+i2] = in->value.rp[k+i2];
      }
      break;
    case APLC_CHAR:
      for (i=0; i<in->size; i++)
	res->value.cp[i] = in->value.cp[i];  
      break;
    case APLC_FNP:
      for (i=0; i<in->size; i++)
	res->value.fnp[i] = in->value.fnp[i];  
      break;
    }
    break;
  }
  res->alloc_type = res->type;
  res->alloc_rank = res->rank;
  res->alloc_size = res->size;
  return;
}

/* sws
   copy an entire trs to another
   - value still points to the original */
extern void
aplc_duptrs(struct trs_struct *res, struct trs_struct *in)
{
  /* type, rank, shape */
  aplc_settrs(res, in->type, in->rank, in->shape);

  if (in->alloc_ind & APLC_ALLOC_VAL_S) {
    res->scalar = in->scalar;
    aplc_link_scalar(res);
  }
  res->value = in->value;
  res->alloc_ind = in->alloc_ind;
  res->alloc_type = in->alloc_type;
  res->alloc_rank = in->alloc_rank;
  res->alloc_size = in->alloc_size;
  return;
}

/* copy a (single) trs value to another trs 
   res(i) = inp(j) */
extern void
aplc_cptrsval(struct trs_struct *res, int i, 
	      struct trs_struct *inp, int j) 
{
  int k, i2,j2;

  /* assume the same type for now... */
  switch(res->type) {
  case APLC_CHAR:
    if (inp->type == APLC_CHAR)
      res->value.cp[i] = inp->value.cp[j];
    else {
      fprintf(aplcerr," char vs %d\n", inp->type);
      aplc_error("[aplc_cptrsval] char type error");
    }
    break;

  case APLC_BOOL:
  case APLC_INT:
  case APLC_LABEL:
    if ( (inp->type == APLC_INT) ||
	 (inp->type == APLC_LABEL) ||
	 (inp->type == APLC_BOOL) )
      res->value.ip[i] = inp->value.ip[j];
    else {
      fprintf(aplcerr," type %d -> int\n", inp->type);
      aplc_error("[aplc_cptrsval] int type error");
    }
    break;

  case APLC_REAL:
    if ( (inp->type == APLC_INT) ||
	 (inp->type == APLC_BOOL) )
      res->value.rp[i] = (double) inp->value.ip[j];
    else if (inp->type == APLC_REAL)
      res->value.rp[i] = inp->value.rp[j];
    else {
      fprintf(aplcerr," type %d -> real\n", inp->type);
      aplc_error("[aplc_cptrsval] real type error");
    }
    break;

  case APLC_COMPLEX:
    i2 = 2*i;
    switch(inp->type) {
    default:
      fprintf(aplcerr," type %d -> complex\n", inp->type);
      aplc_error("[aplc_cptrsval] complex type error");
      break;
    case APLC_BOOL:
    case APLC_INT:
      res->value.rp[i2] = (double) inp->value.ip[j];
      res->value.rp[1+i2] = 0;
      break;
    case APLC_REAL:
      res->value.rp[i2] = inp->value.rp[j];
      res->value.rp[1+i2] = 0;
      break;
    case APLC_COMPLEX:
      j2 = 2*j;
      for (k=0; k<2; k++)
	res->value.rp[k+i2] = inp->value.rp[k+j2];
      break;
    }
    break;

  case APLC_QUAT:
    i2 = 4*i;
    switch(inp->type) {
    default:
      fprintf(aplcerr," type %d -> quat\n", inp->type);
      aplc_error("[aplc_cptrsval] quat type error");
      break;
    case APLC_BOOL:
    case APLC_INT:
      res->value.rp[i2] = (double) inp->value.ip[j];
      for (k=1; k<4; k++)
	res->value.rp[k+i2] = 0;
      break;
    case APLC_REAL:
      res->value.rp[i2] = inp->value.rp[j];
      for (k=1; k<4; k++)
	res->value.rp[k+i2] = 0;
      break;
    case APLC_COMPLEX:
      j2 = 2*j;
      for (k=0; k<2; k++)
	res->value.rp[k+i2] = inp->value.rp[k+j2];
      for (k=2; k<4; k++)
	res->value.rp[k+i2] = 0;
      break;
    case APLC_QUAT:
      j2 = 4*j;
      for (k=0; k<4; k++)
	res->value.rp[k+i2] = inp->value.rp[k+j2];
      break;
    }
    break;

  case APLC_OCT:
    i2 = 8*i;
    switch(inp->type) {
    default:
      fprintf(aplcerr," type %d -> oct\n", inp->type);
      aplc_error("[aplc_cptrsval] oct type error");
      break;
    case APLC_BOOL:
    case APLC_INT:
      res->value.rp[i2] = (double) inp->value.ip[j];
      for (k=1; k<8; k++)
	res->value.rp[k+i2] = 0;
      break;
    case APLC_REAL:
      res->value.rp[i2] = inp->value.rp[j];
      for (k=1; k<8; k++)
	res->value.rp[k+i2] = 0;
      break;
    case APLC_COMPLEX:
      j2 = 2*j; 
      for (k=0; k<2; k++)
	res->value.rp[k+i2] = inp->value.rp[k+j2];
      for (k=2; k<8; k++)
	res->value.rp[k+i2] = 0;
      break;
    case APLC_QUAT:
      j2 = 4*j;
      for (k=0; k<4; k++)
	res->value.rp[k+i2] = inp->value.rp[k+j2];
      for (k=4; k<8; k++)
	res->value.rp[k+i2] = 0;
      break;
    case APLC_OCT:
      j2 = 8*j;
      for (k=0; k<8; k++)
	res->value.rp[k+i2] = inp->value.rp[k+j2];
      break;
    }
    break;

  case APLC_BOXED:
    if (inp->type == APLC_BOXED)
      aplc_copytrs(&res->value.trsp[i], &inp->value.trsp[j]);
    else
      aplc_error("[aplc_cptrsval] boxed type error");
    break;

  case APLC_FNP:
    if (inp->type == APLC_FNP)
      res->value.fnp[i] = inp->value.fnp[j];
    else
      aplc_error("[aplc_cptrsval] function type error");
    break;

  default:
    fprintf(aplcerr, "res type %d\n", res->type);
    aplc_error("[aplc_cptrsval] unknown res type");
  }
  return;
}

/* copy one res to another */ 
extern void
aplc_cpres(res_t *res, res_t *resin, int type)
{
  int j;

  switch (type) {
  case APLC_CHAR:
    res->c = resin->c;
    break;
  case APLC_BOOL:
  case APLC_INT:
    res->i = resin->i;
    break;
  case APLC_REAL:
    res->r = resin->r;
    break;
  case APLC_COMPLEX:
    res->z[0] = resin->z[0];
    res->z[1] = resin->z[1];
    break;
  case APLC_QUAT:
    res->q[0] = resin->q[0];
    res->q[1] = resin->q[1];
    res->q[2] = resin->q[2];
    res->q[3] = resin->q[3];
    break;
  case APLC_OCT:
    for (j=0; j<8; j++)
      res->o[j] = resin->o[j];
    break;
  case APLC_BOXED:
    res->trs = resin->trs;
    break;
  case APLC_FNP:
    res->fn = resin->fn;
    break;
  default:
    fprintf(aplcerr, "type %d\n", type);
    aplc_error("[aplc_cpres] impossible condition");
  }
  return;
}

/* 
   copy a res value to a trs 
   trs_t res(i) = res_t inp
*/
extern void
aplc_cpres2trs(struct trs_struct *res, int i, union res_struct *inp)
{
  int k, i2;

  /* assume the same type for now... */
  switch(res->type) {
  case APLC_CHAR:
    res->value.cp[i] = inp->c;
    break;
  case APLC_BOOL:
  case APLC_INT:
    res->value.ip[i] = inp->i;
    break;
  case APLC_REAL:
    res->value.rp[i] = inp->r;
    break;
  case APLC_COMPLEX:
    i2 = 2*i;
    for (k=0; k<2; k++)
      res->value.rp[k+i2] = inp->z[k];
    break;
  case APLC_QUAT:
    i2 = 4*i;
    for (k=0; k<4; k++)
      res->value.rp[k+i2] = inp->q[k];
    break;
  case APLC_OCT:
    i2 = 8*i;
    for (k=0; k<8; k++)
      res->value.rp[k+i2] = inp->o[k];
    break;
  case APLC_BOXED:
    aplc_copytrs(&res->value.trsp[i], &inp->trs);
    break;
  case APLC_FNP:
    res->value.fnp[i] = inp->fn;
    break;
  default:
    fprintf(aplcerr, "type %d\n", res->type);
    aplc_error("[aplc_cpres2trs] impossible condition");
  }
}


/* sws 
   aplc_settrs - set type, rank, and shape values into a trs structure 

   - note that shape is now copied (sws) rather than setting a pointer 
     so we can memfree (or not) the original source and can later 
     free the trs
   - calculate size  
   - assume this is followed by something like talloc, which allocates
     space for values */
extern void
aplc_settrs_debug(struct trs_struct *trs, 
		  int type, int rank, int *shape,
		  char *file, int line)
{
#if VDEBUG
  printf("[settrs] %s, line %d, type %d, rank %d\n", 
	 file, line, type, rank);
#endif
  aplc_settrs_simple(trs, type, rank, shape);
  return;
}

extern void
aplc_settrs_simple(struct trs_struct * trs, 
		   int type, int rank, int *shape)
{
  int i;
  union mp_struct mptmp;


  trs->type = type;

  /* check to see if space is present already */
  if ( (trs->alloc_ind &  APLC_ALLOC_SHAPE_F) ||
       (trs->alloc_ind &  APLC_ALLOC_SHAPE_NF) ) {
    /* some space for shape is already available
       - is it enough ? */
    if (trs->alloc_rank >= rank) {
      /* have enough */
      trs->rank = rank;
      if (rank > 0) {
        for (i=0; i<rank; i++)
          trs->shape[i] = shape[i];
      }
      /* compute the size */
      trs->size = aplc_vsize(trs->rank, trs->shape);
      return;
    }
    /* else perhaps free 
    if (trs->alloc_ind &  APLC_ALLOC_SHAPE_F)
    aplc_free(trs->shape)
      */
  }

  trs->alloc_ind = APLC_UNALLOC;
  if ( (type == APLC_UNDEF) || 
       (type == APLC_UKTYPE) ) {
    /* ignore request if type doesn't make sense */
    trs->rank = 1;
    trs->shape = aplc_ivzero;
    trs->alloc_rank = 0;
    trs->size = 0;
    return;
  }
  trs->rank = rank;
  trs->alloc_rank = 0;
  if (rank > 0) {
    if ( (rank == 1) && (shape[0] == 0) ) {
      /* zilde - don't allocate, use static 0 */
      trs->shape = aplc_ivzero;
      trs->alloc_ind |= APLC_ALLOC_SHAPE_NF;
    } else {
      aplc_vectalloc(&mptmp, rank, APLC_INT);
      trs->shape = mptmp.ip;
      for (i=0; i<rank; i++)
	trs->shape[i] = shape[i];
      trs->alloc_ind |= APLC_ALLOC_SHAPE_F;
      trs->alloc_rank = rank;
    }
  } else {
    /* rank is zero, must be scalar, put shape=1 anyway */
    trs->shape = aplc_ivone;
    trs->alloc_ind |= APLC_ALLOC_SHAPE_NF;
  }
  /* compute the size */
  trs->size = aplc_vsize(trs->rank, trs->shape);
  return;
}

/* 
   aplc_talloc - allocate storage for values in a trs structure 

   - assume that the rank and shape have been set already
   - assume the size has been determined already
   - return the size 
*/
extern int
aplc_talloc_debug(struct trs_struct *trs, char *file, int line)
{
#if VDEBUG
  printf("[talloc] %s, line %d, type %d, size %d; alloc_type %d, alloc_ind %d \n", 
	 file, line, trs->type, trs->size, trs->alloc_type, trs->alloc_ind);
  aplc_print_alloc_ind(trs->alloc_ind);
#endif
  aplc_talloc_simple(trs);
#if VDEBUG
  printf("[talloc]+ %s, line %d, type %d, size %d, alloc_ind %d \n", 
	 file, line, trs->type, trs->size, trs->alloc_ind);
  aplc_print_alloc_ind(trs->alloc_ind);
#endif
  return (trs->size);
}

extern int
aplc_talloc_simple(struct trs_struct *trs)
{
  /* ignores 0 size requests */
  if (trs->size <1)
    return 0;
#if 1  
  /* to do: 
     - see if we already have enough space
     ...
     if size == 1, then probably should use scalar link anyway...
   */
  if (trs->alloc_ind & APLC_ALLOC_VAL_S) {
    /* value already linked to scalar */
    if ((trs->size == 1) && (trs->type == trs->alloc_type) ) {
#if 0
      /* test/double check */
      aplc_link_scalar(trs);
#endif
      /* we're done */
      return trs->size;
    } else {
      /* can't use current scalar link -- remove */
      RBIT(trs->alloc_ind,APLC_ALLOC_VAL_S);
      RBIT(trs->alloc_ind, APLC_ALLOC_VAL_NF);
#if VDEBUG
      printf("removing old scalar link info\n"); 
      aplc_print_alloc_ind(trs->alloc_ind);
#endif
    }
  }
#endif

  /* if only a singleton, & type ok, don't malloc */
  if (trs->size == 1) {
    switch(trs->type) {
    default:
      break;
    case APLC_BOOL:
    case APLC_INT:
    case APLC_REAL:
    case APLC_CHAR:
    case APLC_COMPLEX:
    case APLC_QUAT:
    case APLC_OCT:
      aplc_link_scalar(trs);
#if VDEBUG
      printf("- making scalar link\n");
#endif
      return trs->size;
      break;
    }
  }

  trs->alloc_ind = 0;/* clear out anything present */
  aplc_vectalloc(&trs->value, trs->size, trs->type);
  trs->alloc_ind |= APLC_ALLOC_VAL_F;
  trs->alloc_type = trs->type;
  trs->alloc_size = trs->size;
  return (trs->size);
}

extern void
aplc_free_debug(void *x, char *file, int line)
{
#if VDEBUG
  /*fprintf(stderr, "[free] %s, line %d, %p\n", file, line, x);*/
  printf("[free] %s, line %d, %p\n", file, line, x);
#endif
#if APLC_MTC
  fprintf(stderr, "[aplc_free_debug] %s, line %d, %p\n", file, line, x);
#endif
#if 0 /* turn off free for debugging ... */
  free(x);
#endif
  return;
}

#if 1
/* probably don't need */
extern void
aplc_free_simple(void *x)
{
#if VDEBUG
  /*fprintf(stderr, "[free] %p\n", x);*/
  printf("[free simple] %p\n", x);
#endif
  free(x);
  return;
}
#endif /* 1 */

/* sws 
   vectalloc - allocate a vector - main memory allocation interface 

   complex: stored as [r i]
*/
extern void
aplc_vectalloc_debug(union mp_struct * mp, int size, int type,
		     char *file, int line)
{
  void *c = NULL;
  int j; 
  int tsize = 0;

#if VDEBUG
  /*fprintf(stderr,
    "[vectalloc_debug] %16s, line %5d, request for size %4d, type %d, ", 
    file, line, size,type);*/
  /*printf("[vectalloc_debug] %16s, line %5d, request for size %4d, type %d, ", 
    file, line, size,type);*/
  printf("[vectalloc_debug] %16s, line %5d, request for size %4d, type %s, ", 
	 file, line, size, str_type_name(type));
#endif

  /*aplc_vectalloc_simple(mp, size, type);*/

  /* quietly ignore requests for no space! */
  /* sws  changed to always give space for 1 */
  if (size <= 0) {
    if (size < 0) {
      fprintf(aplcerr,"[vectalloc_debug] size < 0, forcing crash\n");
      fprintf(NULL,"x");/* force crash so we can stop */
    }
    size = 1;
  }

  switch (type) {
  case APLC_BOXED:
    /* allocate all the pointers */
    tsize = size*sizeof(struct trs_struct);
    c = (void *) APLC_MALLOC(tsize);
    mp->trsp = (struct trs_struct *)c;
    /* also initialize boxed types to null pointers */ 
    for (j=0; j<size; j++)
      aplc_inittrs(&mp->trsp[j]);
    break;

  case APLC_FNP:
    tsize = size*sizeof(aplc_fp_t);
    c = APLC_MALLOC(tsize);
    mp->cp = (char *) c;
    break;

  case APLC_CHAR:
    tsize = size*sizeof(char);
    c = APLC_MALLOC(tsize);
    mp->cp = (char *) c;
    break;

  case APLC_BOOL:
  case APLC_INT:
  case APLC_LABEL:
    /* c = (void *)(mp->ip = APLC_IMALLOC(size));*/
    tsize = size*sizeof(int);
    c = APLC_MALLOC(tsize);
    mp->ip = (int *) c;
    break;

  case APLC_REAL:
    /* c = (void *) (mp->rp = APLC_DMALLOC(size));*/
    tsize = size*sizeof(double);
    c = APLC_MALLOC(tsize);
    mp->rp = (double *) c;
    break;

  case APLC_COMPLEX:
    tsize = 2*size*sizeof(double);
    c = APLC_MALLOC(tsize);
    mp->rp = (double *) c;
    break;

  case APLC_QUAT:
    tsize = 4*size*sizeof(double);
    c = APLC_MALLOC(tsize);
    mp->rp = (double *) c;
    break;

  case APLC_OCT:
    tsize = 8*size*sizeof(double);
    c = APLC_MALLOC(tsize);
    mp->rp = (double *) c;
    break;

  default:
    fprintf(aplcerr, "type %d\n", type);
    aplc_error("[aplc_vectalloc_debug] impossible case");
    break;
  }
  if (c == (void *) 0) {
    fprintf(aplcerr, "request for type %d, size %d\n", type, size);
    aplc_error("[aplc_vectalloc_debug] out of memory allocation space");
  }
#if VDEBUG
  /*fprintf(stderr, " total size %d, %p\n", tsize, c);*/
  printf(" total size %d, %p\n", tsize, c);
#endif
  return;
}

extern void
aplc_vectalloc_simple(union mp_struct * mp, int size, int type)
{
  void *c = (void *) 0;
  int j;

#if VDEBUG
  /*fprintf(stderr,"[vectalloc] request for size %d, type %d\n", size,type);*/
  printf("[vectalloc] request for size %d, type %d\n", size,type);
#endif

  /* error on request for negative space */
  if (size < 0) {
    fprintf(aplcerr, "[vectalloc] request for size %d, type %d\n", size,type);
    aplc_error("[aplc_vectalloc] impossible case");
  }

  /* quietly ignore requests for no space! */
  /* sws  changed to always give space for 1 */
  if (size <= 0) {
#if VDEBUG
    if (size < 0)
      fprintf(NULL,"x");
#endif
    size = 1;
  }

  switch (type) {
  case APLC_BOXED:
    /* allocate all the pointers */
    c = (void *) APLC_MALLOC((size_t) size*sizeof(struct trs_struct));
    mp->trsp = (struct trs_struct *)c;
    /* also initialize boxed types to null pointers */ 
    for (j=0; j<size; j++)
      aplc_inittrs(&mp->trsp[j]);
    break;

  case APLC_FNP:
    c = (void *) APLC_MALLOC((size_t) size*sizeof(aplc_fp_t));
    mp->fnp = (aplc_fp_t *) c;
    break;

  case APLC_CHAR:
    c = APLC_MALLOC(size*sizeof(char));
    mp->cp = (char *) c;
    break;

  case APLC_BOOL:
  case APLC_INT:
    c = (void *)(mp->ip = APLC_IMALLOC(size));
    break;

  case APLC_REAL:
    c = (void *) (mp->rp = APLC_DMALLOC(size));
    break;

  case APLC_COMPLEX:
    c = (void *) (mp->rp = APLC_DMALLOC(2*size));
    break;

  case APLC_QUAT:
    c = (void *) (mp->rp = APLC_DMALLOC(4*size));
    break;

  case APLC_OCT:
    c = (void *) (mp->rp = APLC_DMALLOC(8*size));
    break;

  default:
    fprintf(aplcerr, "type %d\n", type);
    aplc_error("[aplc_vectalloc] impossible case");
    break;
  }
  if (c == (void *) 0) {
    fprintf(aplcerr, "request for type %d, size %d\n", type, size);
    aplc_error("[aplc_vectalloc] out of memory allocation space");
  }
}

/* 
   promote types in a numeric vector 
   - currently only int->real
*/
extern void
aplc_vectpromote(union mp_struct * mp, int size, int type, int newtype )
{
  int i;
  union mp_struct mptemp;

  switch(newtype) {
  default:
  case APLC_CHAR:
  case APLC_COMPLEX:
  case APLC_QUAT:
  case APLC_OCT:
    /* error */
    aplc_error("[aplc_vectpromote] illegal types");
    break;
  case APLC_BOOL:
  case APLC_INT:
    /* nothing to do */ 
    break;
  case APLC_REAL:
    switch(type) {
    default:
      /* error */
      aplc_error("[aplc_vectpromote] illegal types");
      break;
    case APLC_BOOL:
    case APLC_INT:
      /* save old stuff */
      mptemp.ip = mp->ip;
      /* re allocate space */
      aplc_vectalloc(mp, size, APLC_REAL);
      /* copy */
      for (i = size - 1; i >= 0; i--)
	mp->rp[i] = (double) mptemp.ip[i];
      /* free */
      aplc_free(mptemp.ip); 
      break;
    }
    break;
  }
  return;
}

#ifdef SUNOS
/* some special alloc routines to meet sparc alignment requirements 
   - assume this is needed for all sunos machines
   - I don't have any experience with, say, sunos on x86 machines
   - assume this won't hurt
*/

void *
aplc_imalloc(int size)
{
  void *i;
  int nsize;
#if MEMDEBUG
  int in;
  double id;
#endif

  /* adjust size - want size*4 divisible by 8 */
  /* nsize = (8/4)*ceil(size*4/8.0);*/
  /* adjust size - want size*4 divisible by 16 */
  nsize = (16/4)*ceil(size*4/16.0);
#if MEMDEBUG
  printf("[aplc_imalloc] size = %d, nsize = %d\n", size, nsize);  
#endif
  /*i = memalign(8, (unsigned) nsize*sizeof(int));*/
  i = memalign(8, (size_t) nsize*sizeof(int));
#if MEMDEBUG
  in = (int) i;
  id = in/8.0;
  printf("[aplc_imalloc] i = %d, i/8 = %g\n", in, id);  
#endif
  return(i);  
}


/* realloc routines that use memalign, to ensure correct allignment 

   - these assume that allocations come in chunks of APLC_IBUFSIZE
     so if the request is for size, the old size was
     size - APLC_IBUFSIZE
     and both old and new are the same type
*/
extern void *
aplc_drealloc(double *ip, int size)
{
  double *newip, *nipt, *ipt;
  char *c;
  int i, oldsize;

  /* create new storage */
  c = (char *) (newip = (double *) APLC_DMALLOC(size));
  if (c == (char *) 0) {
    fprintf(aplcerr, "request for size %d\n", size);
    aplc_error("[aplc_drealloc] out of memory allocation space");
  }
  /* copy old to new */
  oldsize = size - APLC_IBUFSIZE;
  for (i = 0, ipt = ip, nipt = newip; i < oldsize; i++) {
    *nipt++ = *ipt++;
  }

  /* free old stuff */
  aplc_free(ip);

  return (newip);
}

extern void *
aplc_irealloc(int *ip, int size)
{
  int *newip, *nipt, *ipt;
  char *c;
  int i, oldsize;

  /* create new storage */
  c = (char *) (newip = (int *) APLC_IMALLOC(size));
  if (c == (char *) 0) {
    fprintf(aplcerr, "request for size %d\n", size);
    aplc_error("[aplc_irealloc] out of memory allocation space");
  }
  /* copy old to new */
  oldsize = size - APLC_IBUFSIZE;
  for (i = 0, nipt = newip, ipt = ip; i < oldsize; i++) {
    *nipt++ = *ipt++;
  }

  /* free old stuff */
  aplc_free(ip);

  return (newip);
}

#endif /* #ifdef SUNOS */

/* sws
   vectrealloc - reallocate a vector 
   - main memory allocation interface */
extern void
aplc_vectrealloc(union mp_struct * mp, int size, int type)
{
  void *c = (void *) 0;

  switch (type) {
  case APLC_CHAR:
    /*mp->cp = c = realloc(mp->cp, (unsigned) size);*/
    mp->cp = c = realloc(mp->cp, (size_t) size);
    break;

  case APLC_BOOL:
  case APLC_INT:
    /* mp->ip = (int *) (c = APLC_IREALLOC(size));*/
    c = (char *) (mp->ip = (int *) APLC_IREALLOC(size));
    break;

  case APLC_REAL:
    /* mp->rp = (double *) (c = APLC_DREALLOC(size));*/
    c = (char *) (mp->rp = (double *) APLC_DREALLOC(size));
    break;

  default:
    fprintf(aplcerr, "type %d\n", type);
    aplc_error("[aplc_vectrealloc] impossible case");
    break;
  }
  if (c == (char *) 0) {
    fprintf(aplcerr, "request for type %d, size %d\n", type, size);
    aplc_error("[aplc_vectrealloc] out of memory allocation space");
  }
  return;
}

/* vsize - compute the size of an object
   - product of the shapes, minimum of 1
 */
extern int
aplc_vsize(int rank, int *shape)
{
  register int size;

  size = 1;
  for (rank--; rank >= 0; rank--)
    size *= *(shape + rank);
  return (size);
}


/* sws */
/* detalloc - unallocate storage in a trs structure
              now free both the value and the shape
              also set the pointer values to 0 so no more freeing 
              can happen  */
extern void
aplc_detalloc(struct trs_struct *trs)
{
  int i,n;

  /* ignore requests to free 
     - nill pointers
     - un allocated 
     - not freeable */
  if ( trs == NILTRS) {
#if VDEBUG 
    printf("[aplc_detalloc] arg is nill\n");
#endif
    return;
  }
#if VDEBUG 
  printf("[aplc_detalloc] type %d, rank %d, size %d, alloc %d\n",
	 trs->type, trs->rank, trs->size, trs->alloc_ind);
  if (trs->alloc_ind < 0) {
    fprintf(aplcerr,"trs->alloc_ind < 0, forcing crash\n");
    fprintf(NULL,"x");/* force crash so we can stop */
  }
#endif

#if 0
  /* test case - do nothing */
  return;
#endif

  /* free the shape */
  if ( trs->shape != NULL) {
    if ( !(trs->alloc_ind & APLC_ALLOC_SHAPE_F) ) {
#if VDEBUG 
      printf("[aplc_detalloc] shape is not freeable %d\n", 
	     trs->alloc_ind);
#endif
    } else {
      /* normal vector or array case */
      aplc_free(trs->shape);
    }
  } else {
#if VDEBUG 
    printf("[aplc_detalloc] shape is NILL\n"); 
#endif
  }
  /* force some re-init */
  trs->rank = 0;
  trs->shape = (int *) 0;
  RBIT(trs->alloc_ind, APLC_ALLOC_SHAPE_F);
  RBIT(trs->alloc_ind, APLC_ALLOC_SHAPE_NF);
  trs->alloc_rank = 0;

  /* now free the values */
  if ( trs->alloc_ind & APLC_ALLOC_VAL_F ) { 
    /* watch out for 0 size entities */
    /* get size */
    n = trs->size;
    /*if ( n < 1 )
      return; */
    /* printf("\ntype %d\n", trs->type); */
    switch (trs->type) {
    default:
    case APLC_UNDEF:
    case APLC_UKTYPE:
      fprintf(aplcerr, "[aplc_detalloc] unknown type %d\n", trs->type);
      aplc_error("[aplc_detalloc] impossible case");
      break;
    case APLC_BOOL:
    case APLC_INT:
    case APLC_LABEL:
      aplc_free ( trs->value.ip );
      break;
    case APLC_REAL:
    case APLC_COMPLEX:
    case APLC_QUAT:
    case APLC_OCT:
      aplc_free ( trs->value.rp );
      break;
    case APLC_CHAR:
      aplc_free( trs->value.cp );
      break;
    case APLC_FNP:
      aplc_free( trs->value.fnp );
      break;
    case APLC_BOXED:
      /* recursive */
      for (i=0; i<n; i++)
        aplc_detalloc( &trs->value.trsp[i] );
      aplc_free( trs->value.trsp ); 
      /*fprintf(aplcerr, "[aplc_detalloc] free boxed values?\n");*/
      break;
    }
  } else {
#if VDEBUG 
    printf("[aplc_detalloc] value is  not freeable %d\n", trs->alloc_ind);
#endif
  }

  /* now re-initialize */
  trs->type = APLC_UKTYPE;
  trs->value.ip = (int *) 0;
  trs->alloc_ind = 0;
  trs->alloc_type = APLC_UKTYPE;
  trs->alloc_size = 0;
  return;
}

/* just free the shape of a trs
   - be careful of stuff */
extern void
aplc_free_shape(struct trs_struct * trs)
{
  if (trs == NILTRS) {
#if VDEBUG 
    printf("[aplc_free_shape] trs is nill\n"); 
#endif
    return;
  }
  if (!(trs->alloc_ind & APLC_ALLOC_SHAPE_F)) {
#if VDEBUG 
    printf("[aplc_free_shape] arg is not freeable\n");
#endif
  } else {
    if ( trs->shape != NULL) {
      /* free the shape */
#if VDEBUG 
      printf("[aplc_free_shape] freeing shape vector, length %d\n", trs->rank);
#endif
      aplc_free(trs->shape);
    } else {
#if VDEBUG 
      printf("[aplc_free_shape] shape is NILL\n"); 
#endif
    }
  }
  /* force some re-init */
  trs->rank = 0;
  trs->shape = (int *) 0;
  RBIT(trs->alloc_ind, APLC_ALLOC_SHAPE_F);
  RBIT(trs->alloc_ind, APLC_ALLOC_SHAPE_NF);
  trs->alloc_rank = 0;

#if VDEBUG 
  printf("[aplc_detalloc]+ \n"); 
  aplc_print_trs(trs);
#endif
  return;
}

/* cpvec - copy an integer vector */
extern void
aplc_cpvec(register int *to, int size, register int *from)
{
  for (; size; size--)
    *to++ = *from++;
}

/* cpwo - copy an integer vector without a specific element */
extern void
aplc_cpwo(register int *to, int wo, int size, register int *from)
{
  int i;

  /* fix due to jbww UKC 6/87 */
  for (i = 0; i < size; i++) {
    if (i != wo)
      *to++ = *from;
    from++;
  }
}

/* sws
   cp2str - copy a char vector to a string */
extern void
aplc_cp2str(char *s, struct trs_struct *trs) 
{
  int i, n;
  char *v;

  switch (trs->type) {
  case APLC_CHAR:
    /* n = aplc_vsize(trs->rank, trs->shape);*/
    n = trs->size;
    /*printf("size %d", n);*/
    v = (trs->value).cp;
    for (i=0; i<n; i++)
      s[i] = *v++;
    s[i] = '\0';
    /*printf(" [%s]\n", s);*/
    break;

  default:
    /* error */
    aplc_error("[aplc_cp2str] input not character data");
    break;
  }
}

/* esubi - return the i'th element in the expansion vector 
   - origin 0 */
extern int
aplc_esubi(int i, int rank, int *shape)
{
  register int val;

  val = 1;
  for (rank--; rank > i; rank--)
    val *= *(shape + rank);
  return val;
}

/* generate the whole expansion vector  */
extern int *
aplc_evec(int rank, int *shape)
{
  int *ev;

  if (rank<1)
    rank = 1;/* for a scalar, ensure ev = 1 */
  ev = (int *) APLC_IMALLOC(rank);
  ev[--rank] = 1;
  for (rank--; rank >= 0; rank--)
    ev[rank] = ev[rank+1]*shape[rank+1];
  return ev;
}


#if APLC_MTC
/* sws
   malloc test cases */
void *
aplc_malloc(size_t size, char *fn, int ln)
{
  fprintf(stderr,"[aplc_malloc] %s, line %d,  %ld\n",fn,ln,size);
  return malloc(size);
}

void *
aplc_realloc(void *ptr, size_t size, char *fn, int ln)
{
  fprintf(stderr,"[aplc_realloc] %s, line %d,  %ld\n",fn,ln,size);
  return realloc(ptr,size);
}

#endif /* #if APLC_MTC */

/* end of runmem.c */
